package com.cttic.tcpreceive.init;

import com.cttic.tcpreceive.client.MyClientHandler;
import com.cttic.tcpreceive.client.MyMessageDecoder;
import io.netty.bootstrap.Bootstrap;
import io.netty.channel.Channel;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.codec.string.StringEncoder;
import io.netty.handler.timeout.ReadTimeoutHandler;
import io.netty.handler.timeout.WriteTimeoutHandler;
import java.nio.charset.StandardCharsets;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.stereotype.Component;


@Component
@Slf4j
public class MyClient {

  @Value("${client}")
  private String client;
  @Value("${host}")
  private String host;
  @Value("${port}")
  private Integer port;


  public EventLoopGroup workerGroup;
  private Bootstrap bootstrap;

  public Channel socketChannel;

  public static Boolean CONNECT_STATUS = Boolean.FALSE;

  @Bean
  public void run() throws InterruptedException {
    workerGroup = new NioEventLoopGroup();

    try {
      bootstrap = new Bootstrap();
      socketChannel = new NioSocketChannel();
      bootstrap.group(workerGroup).channel(socketChannel.getClass())
          .handler(new ChannelInitializer<NioSocketChannel>() {
            @Override
            protected void initChannel(NioSocketChannel ch) {
              ChannelPipeline pipeline = ch.pipeline();
              pipeline.addLast(new MyMessageDecoder(Integer.MAX_VALUE, 0, 5));
              pipeline.addLast("encoder", new StringEncoder(StandardCharsets.UTF_8));
              pipeline.addLast(new MyClientHandler());
              pipeline.addLast(new ReadTimeoutHandler(5));
              pipeline.addLast(new WriteTimeoutHandler(5));
            }
          });

      System.out.println("netty client start。。");
      ChannelFuture channelFuture = bootstrap.connect(host, port).sync();
      System.out.println(channelFuture);
//      if (channelFuture.isSuccess()) {
//        log.warn("===   连接Netty服务端成功,connection success");
//        MyClient.CONNECT_STATUS = Boolean.TRUE;
//      }
      channelFuture.channel().writeAndFlush(client);
      channelFuture.channel().closeFuture().sync();
    } finally {
      workerGroup.shutdownGracefully();
    }
  }



}
